import polars as pl
from . import some_cell_series

# register expressions with polars
import h3ronpy.polars as _


def test_expr_cells_resolution():
    df = (
        pl.DataFrame({"cells": some_cell_series()})
        .lazy()
        .with_columns(
            [
                pl.col("cells").h3.cells_resolution().alias("resolution"),
            ]
        )
        .collect()
    )
    assert df["resolution"].dtype == pl.UInt8
    assert df["resolution"][0] == 8


def test_expr_grid_disk():
    df = (
        pl.DataFrame({"cells": some_cell_series()})
        .lazy()
        .with_columns(
            [
                pl.col("cells").h3.grid_disk(1).alias("disk"),
            ]
        )
        .collect()
    )
    assert df["disk"].dtype == pl.List
    assert df["disk"].dtype.inner == pl.UInt64
    assert len(df["disk"][0]) == 7



from h3ronpy.polars import grid_disk, grid_disk_distances, grid_ring_distances, grid_disk_aggregate_k
import numpy as np
import h3.api.numpy_int as h3
import polars as pl


def test_grid_disk():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
            h3.geo_to_h3(5.3, -5.1, 8),
        ],
        dtype=np.uint64,
    )
    disks = grid_disk(h3indexes, 2)
    assert len(disks) == 2
    assert disks.dtype == pl.List(pl.UInt64())

    disks_flat = grid_disk(h3indexes, 2, flatten=True)
    assert len(disks_flat) > 20
    assert disks_flat.dtype == pl.UInt64()


def test_grid_disk_distances():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
            h3.geo_to_h3(5.3, -5.1, 8),
        ],
        dtype=np.uint64,
    )
    disks = grid_disk_distances(h3indexes, 2)
    assert type(disks) == pl.DataFrame
    assert len(disks) == len(h3indexes)
    assert disks["cell"].dtype == pl.List(pl.UInt64())
    assert disks["k"].dtype == pl.List(pl.UInt32())

    centers = (
        grid_disk_distances(h3indexes, 2, flatten=True)
        .lazy()
        .filter(pl.col("cell").is_in(pl.Series(h3indexes)))
        .collect()
    )
    assert len(centers) == len(h3indexes)
    assert len(centers["k"].unique()) == 1
    assert centers["k"].unique()[0] == 0

    # TODO: check values


def test_grid_ring_distances():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
            h3.geo_to_h3(5.3, -5.1, 8),
        ],
        dtype=np.uint64,
    )
    disks = grid_ring_distances(h3indexes, 1, 2)
    assert type(disks) == pl.DataFrame
    assert len(disks) == len(h3indexes)
    assert disks["cell"].dtype == pl.List(pl.UInt64())
    assert disks["k"].dtype == pl.List(pl.UInt32())

    centers = (
        grid_ring_distances(h3indexes, 1, 2, flatten=True)
        .lazy()
        .filter(pl.col("cell").is_in(pl.Series(h3indexes)))
        .collect()
    )
    assert len(centers) == 0

    # TODO: check values


def test_grid_disk_aggregate_k():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
            h3.geo_to_h3(5.3, -5.1, 8),
        ],
        dtype=np.uint64,
    )
    disks = grid_disk_aggregate_k(h3indexes, 2, "max")
    assert type(disks) == pl.DataFrame
    assert disks["cell"].dtype == pl.UInt64()
    assert disks["k"].dtype == pl.UInt32()

    # TODO: check values



from pathlib import Path
import geopandas as gpd

TESTDATA_PATH = Path(__file__).parent.parent / "data"


def load_africa() -> gpd.GeoDataFrame:
    world = gpd.read_file(TESTDATA_PATH / "naturalearth_110m_admin_0_countries.fgb")
    return world[world["CONTINENT"] == "Africa"]



"""
Conversion of 2D `numpy` arrays to H3 cells.

The geo-context is passed to this library using a coordinate transformation matrix - this can be either
a `GDAL-like array <https://gdal.org/tutorials/geotransforms_tut.html>`_ of six float values, or a
`Affine <https://pypi.org/project/affine/>`_-object as used by `rasterio`.

.. note::

    As H3 itself used WGS84 (EPSG:4326) Lat/Lon coordinates, the coordinate transformation matrix used in this module
    must be based on WGS84 as well. Raster data using other coordinate systems need to be reprojected accordingly.


While H3 cells are hexagons and pentagons, this raster conversion process only takes the raster value under the centroid
of the cell into account. When the data shall be aggregated, use any of these methods:

1. Make use the `nearest_h3_resolution` function to convert to the H3 resolution nearest to the pixel size of the raster.
   After that the cell resolution can be changed using the `change_resolution` function and dataframe libraries can be used to
   perform the desired aggregations. This can be a rather memory-intensive process.

2. Scale the raster down using an interpolation algorithm. After that use method 1. This can save a lot of memory, but may
   not be applicable to all datasets - for example dataset with absolute values per pixel like population counts.

Resolution search modes of `nearest_h3_resolution`:

* "min_diff": chose the H3 resolution where the difference in the area of a pixel and the h3index is as small as possible.
* "smaller_than_pixel":  chose the H3 resolution where the area of the h3index is smaller than the area of a pixel.

"""
import shapely
from h3ronpy.h3ronpyrs import raster
from .. import DEFAULT_CELL_COLUMN_NAME
from . import _to_uint64_array, _to_arrow_array
from .vector import cells_to_wkb_polygons, cells_bounds
import numpy as np
import pyarrow as pa
import typing

try:
    # affine library is used by rasterio
    import affine

    __HAS_AFFINE_LIB = True
except ImportError:
    __HAS_AFFINE_LIB = False

Transform = raster.Transform


def _get_transform(t):
    if isinstance(t, Transform):
        return t
    if __HAS_AFFINE_LIB:
        if isinstance(t, affine.Affine):
            return Transform.from_rasterio([t.a, t.b, t.c, t.d, t.e, t.f])
    if type(t) in (list, tuple) and len(t) == 6:
        # probably native gdal
        return Transform.from_gdal(t)
    raise ValueError("unsupported object for transform")


def nearest_h3_resolution(shape, transform, axis_order="yx", search_mode="min_diff") -> int:
    """
    Find the H3 resolution closest to the size of a pixel in an array
    of the given shape with the given transform

    :param shape: dimensions of the 2d array
    :param transform: the affine transformation
    :param axis_order: axis order of the 2d array. Either "xy" or "yx"
    :param search_mode: resolution search mode (see documentation of this module)
    :return:
    """
    return raster.nearest_h3_resolution(shape, _get_transform(transform), axis_order, search_mode)


def raster_to_dataframe(
    in_raster: np.array,
    transform,
    h3_resolution: int,
    nodata_value=None,
    axis_order: str = "yx",
    compact: bool = True,
) -> pa.Table:
    """
    Convert a raster/array to a pandas `DataFrame` containing H3 cell indexes

    This function is parallelized and uses the available CPUs by distributing tiles to a thread pool.

    The input geometry must be in WGS84.

    :param in_raster: Input 2D array
    :param transform:  The affine transformation
    :param nodata_value: The nodata value. For these cells of the array there will be no h3 indexes generated
    :param axis_order: Axis order of the 2d array. Either "xy" or "yx"
    :param h3_resolution: Target h3 resolution
    :param compact: Return compacted h3 indexes (see H3 docs). This results in mixed H3 resolutions, but also can
            reduce the amount of required memory.
    :return: Tuple of arrow arrays
    """

    dtype = in_raster.dtype
    func = None
    if dtype == np.uint8:
        func = raster.raster_to_h3_u8
    elif dtype == np.int8:
        func = raster.raster_to_h3_i8
    elif dtype == np.uint16:
        func = raster.raster_to_h3_u16
    elif dtype == np.int16:
        func = raster.raster_to_h3_i16
    elif dtype == np.uint32:
        func = raster.raster_to_h3_u32
    elif dtype == np.int32:
        func = raster.raster_to_h3_i32
    elif dtype == np.uint64:
        func = raster.raster_to_h3_u64
    elif dtype == np.int64:
        func = raster.raster_to_h3_i64
    elif dtype == np.float32:
        func = raster.raster_to_h3_f32
    elif dtype == np.float64:
        func = raster.raster_to_h3_f64
    else:
        raise NotImplementedError(f"no raster_to_h3 implementation for dtype {dtype.name}")

    return pa.Table.from_arrays(
        arrays=func(in_raster, _get_transform(transform), h3_resolution, axis_order, compact, nodata_value),
        names=["value", DEFAULT_CELL_COLUMN_NAME],
    )


def rasterize_cells(
    cells, values, size: typing.Union[int, typing.Tuple[int, int]], nodata_value=0
) -> (np.ndarray, typing.Tuple[float, float, float, float, float, float]):
    """
    Generate a raster numpy array from arrays of cells and values.

    This function requires the ``rasterio`` library.

    :param cells: array with H3 cells
    :param values: array with the values which shall be written into the raster
    :param size: The desired output size of the raster. Maybe a tuple of ints (width, height) or a single int. In case
            of the latter, the other dimension is interpolated from the bounds of the input data.
    :param nodata_value: The nodata value for the output array
    :return: 2D numpy array typed accordingly to the passed in values array, and the geotransform (WGS84 coordinate
            system, ordering used by the affine library and rasterio)
    """
    from rasterio.transform import from_bounds
    from rasterio.features import rasterize

    cells = _to_uint64_array(cells)
    values = _to_arrow_array(values, None)

    if len(cells) != len(values):
        raise ValueError("length of cells and values array needs to be equal")
    bounds = cells_bounds(cells)

    if bounds is None:
        # no contents, nothing to rasterize
        return None, None

    if type(size) == int:
        (minx, miny, maxx, maxy) = bounds
        size = (size, int(float(size) / (maxx - minx) * (maxy - miny)))

    transform = from_bounds(*bounds, *size)

    # reduce the number of features to loop over by grouping by value
    # https://arrow.apache.org/docs/python/generated/pyarrow.TableGroupBy.html
    grouped = (
        pa.Table.from_arrays([cells, values], names=["cells", "values"])
        .group_by("values")
        .aggregate(
            [
                ("cells", "hash_distinct"),
            ]
        )
    )

    # drop any unused references to free some memory
    del cells
    del values

    values_array = grouped["values"].to_numpy()
    rasterized = np.full(size, nodata_value, dtype=values_array.dtype)

    for cells, value in zip(grouped["cells_distinct"], grouped["values"]):
        value = value.as_py()

        # linking cells should speed up rendering in case of large homogenous areas
        polygons = cells_to_wkb_polygons(cells, link_cells=True)
        polygons = [shapely.from_wkb(polygon.as_py()) for polygon in polygons.filter(polygons.is_valid())]

        # draw
        rasterize(
            polygons,
            out_shape=size,
            out=rasterized,
            default_value=value,
            transform=transform,
            all_touched=False,
        )

    return rasterized, transform



import numpy as np
import math
import h3.api.numpy_int as h3

from h3ronpy.pandas import change_resolution, change_resolution_paired, cells_resolution


def test_change_resolution_up():
    h3indexes = np.array([h3.geo_to_h3(10.2, 45.5, 5), h3.geo_to_h3(10.3, 45.1, 8)], dtype=np.uint64)
    out_res = 9
    changed = change_resolution(h3indexes, out_res)
    assert changed.shape[0] == (int(math.pow(7, 4)) + 7)
    for i in range(len(changed)):
        assert h3.h3_get_resolution(changed[i]) == out_res


def test_change_resolution_paired_up():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
        ],
        dtype=np.uint64,
    )
    out_res = 9
    changed_df = change_resolution_paired(h3indexes, out_res)
    assert len(changed_df) == 7
    for i in range(len(changed_df)):
        assert h3.h3_get_resolution(changed_df["cell_before"][i]) == 8
        assert h3.h3_get_resolution(changed_df["cell_after"][i]) == out_res


def test_change_resolution_down():
    h3indexes = np.array([h3.geo_to_h3(10.2, 45.5, 5), h3.geo_to_h3(10.3, 45.1, 8)], dtype=np.uint64)
    out_res = 4
    changed = change_resolution(h3indexes, out_res)
    assert changed.shape[0] == 2
    assert h3.h3_get_resolution(changed[0]) == out_res
    assert h3.h3_get_resolution(changed[1]) == out_res


def test_cells_resolution():
    h3indexes = np.array([h3.geo_to_h3(10.2, 45.5, 5), h3.geo_to_h3(10.3, 45.1, 8)], dtype=np.uint64)
    res = cells_resolution(h3indexes)
    assert len(res) == 2
    assert res[0] == 5
    assert res[1] == 8



from ..arrow import raster as arrow_raster
from . import _wrap
import polars as pl

nearest_h3_resolution = arrow_raster.nearest_h3_resolution
raster_to_dataframe = _wrap(arrow_raster.raster_to_dataframe, ret_type=pl.DataFrame)
rasterize_cells = arrow_raster.rasterize_cells

__doc__ = arrow_raster.__doc__

__all__ = [nearest_h3_resolution.__name__, raster_to_dataframe.__name__, rasterize_cells.__name__]



import polars as pl
import numpy as np
import h3.api.numpy_int as h3


def some_cell_series() -> pl.Series:
    return pl.Series(
        np.array(
            [
                h3.geo_to_h3(10.3, 45.1, 8),
            ],
            dtype=np.uint64,
        )
    )



try:
    import rasterio

    HAS_RASTERIO = True
except ImportError:
    # rasterio is an optional dependency
    HAS_RASTERIO = False

import numpy as np
import pytest
import polars as pl

from h3ronpy.polars.raster import raster_to_dataframe, rasterize_cells
from h3ronpy import DEFAULT_CELL_COLUMN_NAME, H3_CRS

from tests import TESTDATA_PATH


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_r_tiff():
    dataset = rasterio.open(TESTDATA_PATH / "r.tiff")
    band = dataset.read(1)
    df = raster_to_dataframe(band, dataset.transform, 8, nodata_value=0, compact=True)
    assert len(df) > 100
    assert df[DEFAULT_CELL_COLUMN_NAME].dtype == pl.UInt64
    assert df["value"].dtype == pl.UInt8


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_r_tiff_float32():
    dataset = rasterio.open(TESTDATA_PATH / "r.tiff")
    band = dataset.read(1).astype(np.float32)
    df = raster_to_dataframe(band, dataset.transform, 8, nodata_value=np.NAN, compact=True)
    assert len(df) > 100
    assert df[DEFAULT_CELL_COLUMN_NAME].dtype == pl.UInt64
    assert df["value"].dtype == pl.Float32


def write_gtiff(filename, array, transform, nodata_value):
    with rasterio.open(
        filename,
        mode="w",
        driver="GTiff",
        compress="lzw",
        height=array.shape[0],
        width=array.shape[1],
        count=1,
        dtype=array.dtype,
        crs=H3_CRS,
        transform=transform,
        nodata_value=nodata_value,
    ) as ds:
        ds.write(array, 1)


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_rasterize_cells():
    df = pl.read_parquet(TESTDATA_PATH / "population-841fa8bffffffff.parquet")
    size = (1000, 1000)
    nodata_value = -1
    array, transform = rasterize_cells(df["h3index"], df["pop_general"].cast(pl.Int32), size, nodata_value=nodata_value)

    assert array.shape == size
    assert np.int32 == array.dtype.type
    assert np.any(array > 0)

    # for inspection during debugging
    if False:
        write_gtiff("/tmp/rasterized.tif", array, transform, nodata_value)


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_rasterize_cells_auto_aspect():
    df = pl.read_parquet(TESTDATA_PATH / "population-841fa8bffffffff.parquet")
    size = 1000
    nodata_value = -1
    array, transform = rasterize_cells(df["h3index"], df["pop_general"].cast(pl.Int32), size, nodata_value=nodata_value)

    assert array.shape[0] == size
    # print(array.shape)
    assert np.int32 == array.dtype.type
    assert np.any(array > 0)

    # for inspection during debugging
    if False:
        write_gtiff("/tmp/rasterized_auto_aspect.tif", array, transform, nodata_value)



"""
API to use `h3ronpy` with the `pandas dataframe library <https://pandas.pydata.org>`_ including `geopandas <https://geopandas.org>`_.

.. warning::

    To avoid pulling in unused dependencies, `h3ronpy` does not declare a dependency to `pandas` and `geopandas`. These
    packages need to be installed separately.

"""
from .. import arrow as _arrow
import pyarrow as pa
from functools import wraps
import pandas as pd


def _wrap(func, ret_type=None):
    @wraps(func)
    def wrapper(*args, **kw):
        result = func(*args, **kw)
        if isinstance(result, pa.Table):
            return result.to_pandas(split_blocks=True, self_destruct=True)
        elif isinstance(result, pa.Array):
            return result.to_pandas()
        return result

    if ret_type:
        # create a copy to avoid modifying the dict of the wrapped function
        wrapper.__annotations__ = dict(**wrapper.__annotations__)
        wrapper.__annotations__["return"] = ret_type
    return wrapper


change_resolution = _wrap(_arrow.change_resolution, ret_type=pd.Series)
change_resolution_list = _wrap(_arrow.change_resolution, ret_type=pd.Series)
change_resolution.__annotations__["return"] = pd.Series
change_resolution_paired = _wrap(_arrow.change_resolution_paired, ret_type=pd.DataFrame)
cells_resolution = _wrap(_arrow.cells_resolution, ret_type=pd.Series)
cells_parse = _wrap(_arrow.cells_parse, ret_type=pd.Series)
vertexes_parse = _wrap(_arrow.vertexes_parse, ret_type=pd.Series)
directededges_parse = _wrap(_arrow.directededges_parse, ret_type=pd.Series)
compact = _wrap(_arrow.compact, ret_type=pd.Series)
uncompact = _wrap(_arrow.uncompact, ret_type=pd.Series)
cells_valid = _wrap(_arrow.cells_valid, ret_type=pd.Series)
vertexes_valid = _wrap(_arrow.vertexes_valid, ret_type=pd.Series)
directededges_valid = _wrap(_arrow.directededges_valid, ret_type=pd.Series)
grid_disk = _wrap(_arrow.grid_disk, ret_type=pd.Series)
grid_disk_distances = _wrap(_arrow.grid_disk_distances, ret_type=pd.DataFrame)
grid_ring_distances = _wrap(_arrow.grid_ring_distances, ret_type=pd.DataFrame)
grid_disk_aggregate_k = _wrap(_arrow.grid_disk_aggregate_k, ret_type=pd.DataFrame)
cells_area_m2 = _wrap(_arrow.cells_area_m2, ret_type=pd.Series)
cells_area_km2 = _wrap(_arrow.cells_area_km2, ret_type=pd.Series)
cells_area_rads2 = _wrap(_arrow.cells_area_rads2, ret_type=pd.Series)
cells_to_string = _wrap(_arrow.cells_to_string, ret_type=pd.Series)
vertexes_to_string = _wrap(_arrow.vertexes_to_string, ret_type=pd.Series)
directededges_to_string = _wrap(_arrow.directededges_to_string, ret_type=pd.Series)


__all__ = [
    change_resolution.__name__,
    change_resolution_list.__name__,
    change_resolution_paired.__name__,
    cells_resolution.__name__,
    cells_parse.__name__,
    vertexes_parse.__name__,
    directededges_parse.__name__,
    compact.__name__,
    uncompact.__name__,
    cells_valid.__name__,
    vertexes_valid.__name__,
    directededges_valid.__name__,
    grid_disk.__name__,
    grid_disk_distances.__name__,
    grid_ring_distances.__name__,
    grid_disk_aggregate_k.__name__,
    cells_area_m2.__name__,
    cells_area_km2.__name__,
    cells_area_rads2.__name__,
    cells_to_string.__name__,
    vertexes_to_string.__name__,
    directededges_to_string.__name__,
]



import pytest

from h3ronpy.polars import compact, change_resolution, uncompact
import numpy as np
import h3.api.numpy_int as h3


def compact_to_one(expected_cell, input_cells, **kw):
    compacted = compact(input_cells, **kw)
    assert len(compacted) == 1
    assert compacted[0] == expected_cell


def test_compact():
    cell = h3.geo_to_h3(10.3, 45.1, 8)
    h3indexes = change_resolution(
        np.array(
            [
                cell,
            ],
            dtype=np.uint64,
        ),
        10,
    )
    compact_to_one(cell, h3indexes)


def test_compact_mixed_fail():
    cell = h3.geo_to_h3(10.3, 45.1, 8)
    with pytest.raises(ValueError, match="heterogen"):
        compact_to_one(cell, [cell, h3.h3_to_parent(cell, 4)])


def test_compact_mixed():
    cell = h3.geo_to_h3(10.3, 45.1, 8)
    compact_to_one(cell, [cell, h3.geo_to_h3(10.3, 45.1, 9)], mixed_resolutions=True)


def test_uncompact():
    cells = uncompact(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
        ],
        9,
    )
    assert len(cells) == 7



# install the dependencies needed for development and ci by
# collecting them from all relevant files

import subprocess
from pathlib import Path

import sys
import os


def install(packages, upgrade=False):
    pkg_manager = os.environ.get("PKG_MANAGER") or "pip"
    cmd = [pkg_manager, "install"]
    if upgrade and pkg_manager == "pip":
        cmd.append("--upgrade")
    if packages:
        subprocess.run(cmd + packages, stdout=sys.stdout, stderr=sys.stderr)


if __name__ == "__main__":
    # do not update pip - may fail on windows
    install(["toml", "black", "ruff"], upgrade=True)  # always upgrade pip

    packages = []

    if sys.platform == "linux":
        packages.append("patchelf")

    def harvest_deps(section, keys):
        for k in keys:
            packages.extend(section.get(k, []))

    import toml  # import only after it has been installed

    pyproject_toml = toml.load(Path(__file__).parent / "pyproject.toml")
    harvest_deps(pyproject_toml.get("build-system", {}), ("requires", "requires-dist"))
    project = pyproject_toml.get("project", {})
    harvest_deps(project, ("dependencies",))

    for deps in project.get("optional-dependencies", {}).values():
        packages.extend(deps)

    pytest = pyproject_toml.get("tool", {}).get("pytest")
    if pytest is not None:
        pytest_package = "pytest"
        pytest_minversion = pytest.get("ini_options", {}).get("minversion")
        if pytest_minversion:
            packages.append(f"{pytest_package}>={pytest_minversion}")
        else:
            packages.append(f"{pytest_package}")

    install(packages)



from h3ronpy.h3ronpyrs import vector
from .. import ContainmentMode
from . import _to_uint64_array, _HAS_POLARS, _to_arrow_array
from typing import Optional, Tuple, Union
import pyarrow as pa


def cells_to_coordinates(arr, radians: bool = False) -> pa.Table:
    """
    convert to point coordinates in degrees
    """
    return vector.cells_to_coordinates(_to_uint64_array(arr), radians=radians)


def coordinates_to_cells(latarray, lngarray, resarray, radians: bool = False) -> pa.Array:
    """
    Convert coordinates arrays to cells.

    :param latarray: array of lat values
    :param lngarray: array of lng values
    :param resarray: Either an array of resolutions or a single resolution as an integer to apply to all coordinates.
    :param radians: Set to True to pass `lat` and `lng` in radians
    :return: cell array
    """
    if type(resarray) in (int, float):
        res = int(resarray)
    else:
        res = _to_arrow_array(resarray, pa.uint8())
    return vector.coordinates_to_cells(
        _to_arrow_array(latarray, pa.float64()), _to_arrow_array(lngarray, pa.float64()), res, radians=radians
    )


def cells_bounds(arr) -> Optional[Tuple]:
    """
    Bounds of the complete array as a tuple `(minx, miny, maxx, maxy)`.
    """
    return vector.cells_bounds(_to_uint64_array(arr))


def cells_bounds_arrays(arr) -> pa.Table:
    """
    Build a table/dataframe with the columns `minx`, `miny`, `maxx` and `maxy` containing the bounds of the individual
    cells from the input array.
    """
    return vector.cells_bounds_arrays(_to_uint64_array(arr))


def cells_to_wkb_polygons(arr, radians: bool = False, link_cells: bool = False) -> pa.Array:
    """
    Convert cells to polygons.

    The returned geometries in the output array will match the order of the input array - unless ``link_cells``
    is set to True.

    :param: arr: The cell array
    :param radians: Generate geometries using radians instead of degrees
    :param link_cells: Combine neighboring cells into a single polygon geometry.
    """
    return vector.cells_to_wkb_polygons(_to_uint64_array(arr), radians=radians, link_cells=link_cells)


def cells_to_wkb_points(arr, radians: bool = False) -> pa.Array:
    """
    Convert cells to points using their centroids.

    The returned geometries in the output array will match the order of the input array.

    :param: arr: The cell array
    :param radians: Generate geometries using radians instead of degrees
    """
    return vector.cells_to_wkb_points(_to_uint64_array(arr), radians=radians)


def vertexes_to_wkb_points(arr, radians: bool = False) -> pa.Array:
    """
    Convert vertexes to points.

    The returned geometries in the output array will match the order of the input array.

    :param: arr: The vertex array
    :param radians: Generate geometries using radians instead of degrees
    """
    return vector.vertexes_to_wkb_points(_to_uint64_array(arr), radians=radians)


def directededges_to_wkb_linestrings(arr, radians: bool = False) -> pa.Array:
    """
    Convert directed edges to linestrings.

    The returned geometries in the output array will match the order of the input array.

    :param: arr: The directed edge array
    :param radians: Generate geometries using radians instead of degrees
    """
    return vector.directededges_to_wkb_linestrings(_to_uint64_array(arr), radians=radians)


def wkb_to_cells(
    arr,
    resolution: int,
    containment_mode: ContainmentMode = ContainmentMode.ContainsCentroid,
    compact: bool = False,
    flatten: bool = False,
) -> Union[pa.Array, pa.ListArray]:
    """
    Convert a Series/Array/List of WKB values to H3 cells.

    Unless ``flatten`` is set to True a list array will be returned, with the cells generated from a geometry being
    located at the same position as the geometry in the input array.

    :param arr: The input array.
    :param resolution: H3 resolution
    :param containment_mode: Containment mode used to decide if a cell is contained in a polygon or not.
            See the ContainmentMode class.
    :param compact: Compact the returned cells by replacing cells with their parent cells when all children
            of that cell are part of the set.
    :param flatten: Return a non-nested cell array instead of a list array.
    """
    if _HAS_POLARS:
        import polars as pl

        if isinstance(arr, pl.Series):
            arr = arr.to_arrow()

    if not isinstance(arr, pa.LargeBinaryArray):
        arr = pa.array(arr, type=pa.large_binary())

    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()
    return vector.wkb_to_cells(
        arr,
        resolution,
        containment_mode=containment_mode,
        compact=compact,
        flatten=flatten,
    )


def geometry_to_cells(
    geom,
    resolution: int,
    containment_mode: ContainmentMode = ContainmentMode.ContainsCentroid,
    compact: bool = False,
) -> pa.Array:
    """
    Convert a single object which supports the python `__geo_interface__` protocol to H3 cells

    :param geom: geometry
    :param resolution: H3 resolution
    :param containment_mode: Containment mode used to decide if a cell is contained in a polygon or not.
            See the ContainmentMode class.
    :param compact: Compact the returned cells by replacing cells with their parent cells when all children
            of that cell are part of the set.
    """
    return vector.geometry_to_cells(geom, resolution, containment_mode=containment_mode, compact=compact)


__all__ = [
    cells_to_coordinates.__name__,
    coordinates_to_cells.__name__,
    cells_bounds.__name__,
    cells_bounds_arrays.__name__,
    cells_to_wkb_polygons.__name__,
    cells_to_wkb_points.__name__,
    vertexes_to_wkb_points.__name__,
    directededges_to_wkb_linestrings.__name__,
    wkb_to_cells.__name__,
    geometry_to_cells.__name__,
]



import pyarrow.compute as pc
import pyarrow as pa


# from https://issues.apache.org/jira/browse/ARROW-12099
def explode_table_include_null(table: pa.Table, column: str) -> pa.Table:
    other_columns = list(table.schema.names)
    other_columns.remove(column)
    indices = pc.list_parent_indices(pc.fill_null(table[column], [None]))
    result = table.select(other_columns)
    try:
        # may result in a large memory allocation
        result = result.take(indices)
    except pa.ArrowIndexError:
        # See https://github.com/nmandery/h3ronpy/issues/40
        # Using RuntimeWarning as ResourceWarning is often not displayed to the user.
        import warnings

        warnings.warn("This ArrowIndexError may be a sign of the process running out of memory.", RuntimeWarning)
        raise
    result = result.append_column(
        pa.field(column, table.schema.field(column).type.value_type),
        pc.list_flatten(pc.fill_null(table[column], [None])),
    )
    return result



import shapely

import pandas as pd
from shapely.geometry import Point, GeometryCollection, Polygon
import pytest
from h3ronpy.pandas import change_resolution
from h3ronpy.pandas.vector import (
    cells_to_points,
    cells_to_polygons,
    cells_dataframe_to_geodataframe,
    geodataframe_to_cells,
    geoseries_to_cells,
)
from h3ronpy import DEFAULT_CELL_COLUMN_NAME, ContainmentMode
import geopandas as gpd
from .. import load_africa, TESTDATA_PATH


def test_cells_to_points():
    gs = cells_to_points(
        [
            0x8009FFFFFFFFFFF,
        ]
    )
    assert isinstance(gs, gpd.GeoSeries)
    assert gs.geom_type[0] == "Point"


def test_cells_to_polygons():
    cells = change_resolution(
        [
            0x8009FFFFFFFFFFF,
        ],
        3,
    )
    gs = cells_to_polygons(cells)
    assert isinstance(gs, gpd.GeoSeries)
    assert gs.geom_type[0] == "Polygon"
    assert len(gs) == 286

    linked_gs = cells_to_polygons(cells, link_cells=True)
    assert isinstance(linked_gs, gpd.GeoSeries)
    assert linked_gs.geom_type[0] == "Polygon"
    assert len(linked_gs) == 1
    assert shapely.get_num_coordinates(linked_gs[0]) > 120


def test_cells_dataframe_to_geodataframe():
    df = pd.DataFrame(
        {
            DEFAULT_CELL_COLUMN_NAME: [
                0x8009FFFFFFFFFFF,
            ],
            "id": [
                5,
            ],
        }
    )
    gdf = cells_dataframe_to_geodataframe(df)
    assert isinstance(gdf, gpd.GeoDataFrame)
    assert len(gdf) == len(df)
    assert (gdf[DEFAULT_CELL_COLUMN_NAME] == df[DEFAULT_CELL_COLUMN_NAME]).all()
    assert (gdf["id"] == df["id"]).all()
    assert gdf.geometry.geom_type[0] == "Polygon"


def test_cells_dataframe_to_geodataframe_empty():
    # https://github.com/nmandery/h3ron/issues/17
    df = pd.DataFrame({DEFAULT_CELL_COLUMN_NAME: []})
    gdf = cells_dataframe_to_geodataframe(df)  # should not raise an ValueError.
    assert gdf.empty


def test_cells_geodataframe_to_cells():
    africa = load_africa()
    df = geodataframe_to_cells(africa, 4)
    assert len(df) > len(africa)
    assert df.dtypes[DEFAULT_CELL_COLUMN_NAME] == "uint64"


def test_geoseries_to_cells_flatten():
    africa = load_africa()
    cells = geoseries_to_cells(africa.geometry, 4, flatten=True)
    assert len(cells) >= len(africa)
    assert cells.dtype == "uint64"


@pytest.mark.skip(
    reason="GeometryCollections are unsupported until https://github.com/geoarrow/geoarrow-rs/blob/3a2aaa883126274037cabaf46b1f5f6459938297/src/io/wkb/reader/geometry_collection.rs#L23 is fixed"
)
def test_empty_geometrycollection_omitted():
    gdf = gpd.GeoDataFrame(
        {
            "geometry": [
                GeometryCollection(),
            ]
        },
        crs="epsg:4326",
    )
    df = geodataframe_to_cells(gdf, 4)
    assert len(df) == 0


def test_fail_on_empty_point():
    gdf = gpd.GeoDataFrame(
        {
            "geometry": [
                Point(),
            ]
        },
        crs="epsg:4326",
    )
    with pytest.raises(ValueError):
        geodataframe_to_cells(gdf, 4)


def test_geometry_results_in_no_cells():
    gdf = gpd.GeoDataFrame(
        {
            "geometry": [
                Polygon(
                    [
                        (1.100000, 4.50000),
                        (1.100001, 4.50000),
                        (1.100001, 4.50001),
                        (1.100000, 4.50001),
                        (1.100000, 4.50000),
                    ]
                ),
            ],
            "col1": [1],
        },
        crs="epsg:4326",
    )
    df = geodataframe_to_cells(gdf, 4, containment_mode=ContainmentMode.ContainsCentroid)
    assert len(df) == 0


def test_non_standard_geometry_column_name():
    africa = load_africa()
    assert africa.geometry.name == "geometry"
    africa.rename_geometry("something_else", inplace=True)
    assert africa.geometry.name == "something_else"
    df = geodataframe_to_cells(africa, 4)
    assert len(df) > len(africa)
    assert df.dtypes[DEFAULT_CELL_COLUMN_NAME] == "uint64"



from . import h3ronpyrs as _native
from .h3ronpyrs import version, DEFAULT_CELL_COLUMN_NAME, ContainmentMode  # noqa: F401

__version__ = version()

H3_CRS = "EPSG:4326"

if not _native.is_release_build():
    import warnings

    warnings.warn("h3ronpy has not been compiled in release mode. Performance will be degraded.", RuntimeWarning)


__all__ = [
    "H3_CRS",
    "DEFAULT_CELL_COLUMN_NAME",
    ContainmentMode.__name__,
    version.__name__,
]



from typing import Union

import pyarrow as pa
from h3ronpy.h3ronpyrs import op

try:
    import polars as pl

    _HAS_POLARS = True
except ImportError:
    _HAS_POLARS = False


def _to_arrow_array(arr, dtype) -> pa.Array:
    converted = None
    if _HAS_POLARS:
        if isinstance(arr, pl.Series):
            converted = arr.to_arrow()

    if converted is None:
        converted = pa.array(arr, type=dtype)

    if isinstance(arr, pa.ChunkedArray):
        converted = converted.combine_chunks()
    return converted


def _to_uint64_array(arr) -> pa.Array:
    return _to_arrow_array(arr, pa.uint64())


def change_resolution(arr, resolution: int) -> pa.Array:
    """
    Change the H3 resolutions of all contained values to `resolution`.

    In case of resolution increases all child indexes will be added, so the returned
    value may contain more indexes than the input.

    Invalid/empty values are omitted.
    """
    return op.change_resolution(_to_uint64_array(arr), resolution)


def change_resolution_list(arr, resolution: int) -> pa.Array:
    """
    Change the H3 resolutions of all contained values to `resolution`.

    The output list array has the same length as the input array, positions of the elements
    in input and output are corresponding to each other.

    Invalid/empty values are preserved as such.
    """
    return op.change_resolution_list(_to_uint64_array(arr), resolution)


def change_resolution_paired(arr, resolution: int) -> pa.Table:
    """
    Returns a table/dataframe with two columns: `cell_before` and `cell_after`
    with the cells h3index before and after the resolution change.

    This can be helpful when joining data in different resolutions via
    dataframe libraries
    """
    return op.change_resolution_paired(_to_uint64_array(arr), resolution)


def cells_resolution(arr) -> pa.Array:
    """
    Generates a new array containing the resolution of each cell of the
    input array.

    :param arr:
    :return:
    """
    return op.cells_resolution(_to_uint64_array(arr))


def cells_parse(arr, set_failing_to_invalid: bool = False) -> pa.Array:
    """
    Parse H3 cells from string arrays.

    Setting `set_failing_to_invalid` to true will trigger setting the validity bitmap according
    the successful parsing of an individual element. Having this set to false will cause the
    method to fail upon encountering the first unparsable value.

    This function is able to parse multiple representations of H3 cells:

        * hexadecimal (Example: ``8552dc63fffffff``)
        * numeric integer strings (Example: ``600436454824345599``)
        * strings like ``[x], [y], [resolution]`` or  ``[x]; [y]; [resolution]``. (Example: ``10.2,45.5,5``)
    """
    return op.cells_parse(_to_arrow_array(arr, pa.utf8()), set_failing_to_invalid=set_failing_to_invalid)


def vertexes_parse(arr, set_failing_to_invalid: bool = False) -> pa.Array:
    """
    Parse H3 vertexes from string arrays.

    Setting `set_failing_to_invalid` to true will trigger setting the validity bitmap according
    the successful parsing of an individual element. Having this set to false will cause the
    method to fail upon encountering the first unparsable value.
    """
    return op.vertexes_parse(_to_arrow_array(arr, pa.utf8()), set_failing_to_invalid=set_failing_to_invalid)


def directededges_parse(arr, set_failing_to_invalid: bool = False) -> pa.Array:
    """
    Parse H3 directed edges from string arrays.

    Setting `set_failing_to_invalid` to true will trigger setting the validity bitmap according
    the successful parsing of an individual element. Having this set to false will cause the
    method to fail upon encountering the first unparsable value.
    """
    return op.directededges_parse(_to_arrow_array(arr, pa.utf8()), set_failing_to_invalid=set_failing_to_invalid)


def compact(arr, mixed_resolutions: bool = False) -> pa.Array:
    """
    Compact the given cells

    The cells are expected to be of the same resolution, otherwise this operation will fail unless
    `mixed_resolutions` is set to True. Setting this may lead to slight slow-downs.
    """
    return op.compact(_to_uint64_array(arr), mixed_resolutions=mixed_resolutions)


def uncompact(arr, target_resolution: int) -> pa.Array:
    """
    Uncompact the given cells to the resolution `target_resolution`.

    All higher resolution cells contained in the input array than the given `target_resolution` will
    be omitted from the output.
    """
    return op.uncompact(_to_uint64_array(arr), target_resolution)


def _make_h3index_valid_wrapper(fn, h3index_name, wrapper_name):
    def valid_wrapper(arr, booleanarray: bool = False) -> pa.Array:
        return fn(_to_uint64_array(arr), booleanarray=booleanarray)

    valid_wrapper.__doc__ = f"""
    Validate an array of potentially invalid {h3index_name} values by returning a new
    UInt64 array with the validity mask set accordingly.
    
    If `booleanarray` is set to True, a boolean array describing the validity will be
    returned instead.
    """
    valid_wrapper.__name__ = wrapper_name
    return valid_wrapper


cells_valid = _make_h3index_valid_wrapper(op.cells_valid, "cell", "cells_valid")
vertexes_valid = _make_h3index_valid_wrapper(op.cells_valid, "vertex", "vertexes_valid")
directededges_valid = _make_h3index_valid_wrapper(op.cells_valid, "directed edge", "directededges_valid")


def grid_disk(cellarray, k: int, flatten: bool = False) -> Union[pa.ListArray, pa.Array]:
    return op.grid_disk(_to_uint64_array(cellarray), k, flatten=flatten)


def grid_disk_distances(cellarray, k: int, flatten: bool = False) -> pa.Table:
    return op.grid_disk_distances(_to_uint64_array(cellarray), k, flatten=flatten)


def grid_disk_aggregate_k(cellarray, k: int, aggregation_method: str) -> pa.Table:
    """
    Valid values for `aggregation_method` are `"min"` and `"max"`.
    """
    return op.grid_disk_aggregate_k(_to_uint64_array(cellarray), k, aggregation_method)


def grid_ring_distances(cellarray, k_min: int, k_max: int, flatten: bool = False) -> pa.Table:
    return op.grid_ring_distances(_to_uint64_array(cellarray), k_min, k_max, flatten=flatten)


def cells_area_m2(cellarray) -> pa.Array:
    return op.cells_area_m2(_to_uint64_array(cellarray))


def cells_area_km2(cellarray) -> pa.Array:
    return op.cells_area_km2(_to_uint64_array(cellarray))


def cells_area_rads2(cellarray) -> pa.Array:
    return op.cells_area_rads2(_to_uint64_array(cellarray))


def cells_to_string(cellarray) -> pa.Array:
    return op.cells_to_string(_to_uint64_array(cellarray))


def vertexes_to_string(vertexesarray) -> pa.Array:
    return op.vertexes_to_string(_to_uint64_array(vertexesarray))


def directededges_to_string(directededgearray) -> pa.Array:
    return op.directededges_to_string(_to_uint64_array(directededgearray))


__all__ = [
    change_resolution.__name__,
    change_resolution_list.__name__,
    change_resolution_paired.__name__,
    cells_resolution.__name__,
    cells_parse.__name__,
    vertexes_parse.__name__,
    directededges_parse.__name__,
    compact.__name__,
    uncompact.__name__,
    cells_valid.__name__,
    vertexes_valid.__name__,
    directededges_valid.__name__,
    grid_disk.__name__,
    grid_disk_distances.__name__,
    grid_ring_distances.__name__,
    grid_disk_aggregate_k.__name__,
    cells_area_m2.__name__,
    cells_area_km2.__name__,
    cells_area_rads2.__name__,
    cells_to_string.__name__,
    vertexes_to_string.__name__,
    directededges_to_string.__name__,
]



import polars as pl
from . import some_cell_series

# register expressions with polars
import h3ronpy.polars as _


def test_series_cells_resolution():
    resolution = some_cell_series().h3.cells_resolution()
    assert resolution.dtype == pl.UInt8
    assert resolution[0] == 8



from . import _wrap
from ..arrow import vector as _av
from .. import ContainmentMode
from ..arrow import util as _arrow_util
import pyarrow as pa
import pandas as pd
import geopandas as gpd
from functools import wraps
from typing import Optional
from .. import H3_CRS, DEFAULT_CELL_COLUMN_NAME


def _geoseries_from_wkb(func, doc: Optional[str] = None, name: Optional[str] = None):
    @wraps(func)
    def wrapper(*args, **kw):
        return gpd.GeoSeries.from_wkb(func(*args, **kw), crs=H3_CRS)

    # create a copy to avoid modifying the dict of the wrapped function
    wrapper.__annotations__ = dict(**wrapper.__annotations__)
    wrapper.__annotations__["return"] = gpd.GeoSeries
    if doc is not None:
        wrapper.__doc__ = doc
    if name is not None:
        wrapper.__name__ = name

    return wrapper


cells_to_coordinates = _wrap(_av.cells_to_coordinates, ret_type=pd.DataFrame)
coordinates_to_cells = _wrap(_av.coordinates_to_cells, ret_type=pd.Series)
cells_bounds = _av.cells_bounds
cells_bounds_arrays = _wrap(_av.cells_bounds_arrays, ret_type=pd.DataFrame)
cells_to_wkb_polygons = _wrap(_av.cells_to_wkb_polygons, ret_type=pd.Series)
cells_to_polygons = _geoseries_from_wkb(
    cells_to_wkb_polygons,
    doc="Create a geoseries containing the polygon geometries of a cell array",
    name="cells_to_polygons",
)
cells_to_wkb_points = _wrap(_av.cells_to_wkb_points, ret_type=pd.Series)
cells_to_points = _geoseries_from_wkb(
    cells_to_wkb_points,
    doc="Create a geoseries containing the centroid point geometries of a cell array",
    name="cells_to_points",
)
vertexes_to_wkb_points = _wrap(_av.vertexes_to_wkb_points, ret_type=pd.Series)
vertexes_to_points = _geoseries_from_wkb(
    vertexes_to_wkb_points,
    doc="Create a geoseries containing the point geometries of a vertex array",
    name="vertexes_to_points",
)
directededges_to_wkb_linestrings = _wrap(_av.directededges_to_wkb_linestrings, ret_type=pd.Series)
directededges_to_linestrings = _geoseries_from_wkb(
    directededges_to_wkb_linestrings,
    doc="Create a geoseries containing the linestrings geometries of a directededge array",
    name="directededges_to_linestrings",
)
wkb_to_cells = _wrap(_av.wkb_to_cells, ret_type=pd.Series)
geometry_to_cells = _wrap(_av.geometry_to_cells, ret_type=pd.Series)


@wraps(wkb_to_cells)
def geoseries_to_cells(geoseries: gpd.GeoSeries, *args, **kw):
    return _av.wkb_to_cells(geoseries.to_wkb(), *args, **kw).to_pandas()


geoseries_to_cells.__name__ = "geoseries_to_cells"


def cells_dataframe_to_geodataframe(
    df: pd.DataFrame, cell_column_name: str = DEFAULT_CELL_COLUMN_NAME
) -> gpd.GeoDataFrame:
    """
    Convert a dataframe with a column containing cells to a geodataframe

    :param df: input dataframe
    :param cell_column_name: name of the column containing the h3 indexes
    :return: GeoDataFrame
    """
    return gpd.GeoDataFrame(df, geometry=cells_to_polygons(df[cell_column_name]), crs=H3_CRS)


def geodataframe_to_cells(
    gdf: gpd.GeoDataFrame,
    resolution: int,
    containment_mode: ContainmentMode = ContainmentMode.ContainsCentroid,
    compact: bool = False,
    cell_column_name: str = DEFAULT_CELL_COLUMN_NAME,
) -> pd.DataFrame:
    """
    Convert a `GeoDataFrame` to H3 cells while exploding all other columns according to the number of cells derived
    from the rows geometry.

    The conversion of GeoDataFrames is parallelized using the available CPUs.

    The duplication of all non-cell columns leads to increased memory requirements. Depending on the use-case
    some of the more low-level conversion functions should be preferred.

    :param gdf:
    :param resolution: H3 resolution
    :param containment_mode: Containment mode used to decide if a cell is contained in a polygon or not.
            See the ContainmentMode class.
    :param compact: Compact the returned cells by replacing cells with their parent cells when all children
            of that cell are part of the set.
    :param cell_column_name:
    :return:
    """
    cells = _av.wkb_to_cells(
        gdf.geometry.to_wkb(),
        resolution,
        containment_mode=containment_mode,
        compact=compact,
        flatten=False,
    )
    table = pa.Table.from_pandas(pd.DataFrame(gdf.drop(columns=gdf.geometry.name))).append_column(
        cell_column_name, cells
    )
    return _arrow_util.explode_table_include_null(table, cell_column_name).to_pandas().reset_index(drop=True)


__all__ = [
    cells_to_coordinates.__name__,
    coordinates_to_cells.__name__,
    cells_bounds.__name__,
    cells_bounds_arrays.__name__,
    cells_to_wkb_polygons.__name__,
    cells_to_polygons.__name__,
    cells_to_wkb_points.__name__,
    cells_to_points.__name__,
    vertexes_to_wkb_points.__name__,
    vertexes_to_points.__name__,
    directededges_to_wkb_linestrings.__name__,
    directededges_to_linestrings.__name__,
    cells_dataframe_to_geodataframe.__name__,
    wkb_to_cells.__name__,
    geometry_to_cells.__name__,
    geoseries_to_cells.__name__,
    geodataframe_to_cells.__name__,
]



import pytest

from h3ronpy.polars import cells_parse, cells_valid, cells_to_string
import numpy as np
import h3.api.numpy_int as h3
import polars as pl


def test_cells_parse():
    strings = np.array([h3.h3_to_string(h3.geo_to_h3(45.5, 10.2, 5)), "10.2, 45.5, 5"])
    cells = cells_parse(strings)
    assert len(cells) == 2
    assert cells[0] == cells[1]


def test_cells_parse_largeutf8():
    # polars uses LargeUtf8 datatype for strings
    cells = cells_parse(pl.Series(["801ffffffffffff"]))
    assert len(cells) == 1


def test_parse_cell_fail():
    strings = np.array(
        [
            "invalid",
        ]
    )
    with pytest.raises(ValueError, match="non-parsable CellIndex"):
        cells_parse(strings)


def test_parse_cell_set_invalid():
    strings = np.array(
        [
            "invalid",
        ]
    )
    cells = cells_parse(strings, set_failing_to_invalid=True)
    assert len(cells) == 1
    assert cells[0] is None


def test_cells_valid():
    input = np.array(
        [45, h3.geo_to_h3(45.5, 10.2, 5)],
        dtype=np.uint64,
    )
    cells = cells_valid(input, booleanarray=False)
    assert len(cells) == 2
    assert cells.dtype == pl.datatypes.UInt64()
    assert cells[0] is None
    assert cells[1] is not None

    bools = cells_valid(input, booleanarray=True)
    assert len(bools) == 2
    assert bools.dtype == pl.datatypes.Boolean()
    assert bools[0] is False
    assert bools[1] is True

    assert cells.is_not_null().eq(bools).all()


def test_cells_to_string():
    cells = np.array(
        [
            h3.geo_to_h3(45.5, 10.2, 5),
        ]
    )
    strings = cells_to_string(cells)
    assert len(strings) == len(cells)
    assert isinstance(strings, pl.Series)
    assert strings.dtype == pl.Utf8
    assert strings[0] == "851f9923fffffff"



from h3ronpy.polars.vector import cells_to_coordinates, cells_bounds, cells_bounds_arrays, coordinates_to_cells
import polars as pl
import numpy as np
import h3.api.numpy_int as h3


def test_cells_to_coordinates():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
        ],
        dtype=np.uint64,
    )
    coords = cells_to_coordinates(h3indexes)
    assert len(coords) == 1
    assert 10.0 < coords["lat"][0] < 11.0
    assert 45.0 < coords["lng"][0] < 46.0


def test_coordinates_to_cells():
    lat = np.array([10.3, 23.1], dtype=np.float64)
    lng = np.array([45.1, 2.3], dtype=np.float64)
    r = 7
    cells = coordinates_to_cells(lat, lng, r)
    assert len(cells) == 2
    assert cells[0] == h3.geo_to_h3(lat[0], lng[0], r)
    assert cells[1] == h3.geo_to_h3(lat[1], lng[1], r)


def test_coordinates_to_cells_resolution_array():
    lat = np.array([10.3, 23.1], dtype=np.float64)
    lng = np.array([45.1, 2.3], dtype=np.float64)
    r = np.array([9, 12], dtype=np.uint8)
    cells = coordinates_to_cells(lat, lng, r)
    assert len(cells) == 2
    assert cells[0] == h3.geo_to_h3(lat[0], lng[0], r[0])
    assert cells[1] == h3.geo_to_h3(lat[1], lng[1], r[1])


def test_cells_bounds():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
        ],
        dtype=np.uint64,
    )
    bounds = cells_bounds(h3indexes)
    assert bounds is not None
    assert type(bounds) == tuple
    assert len(bounds) == 4
    assert bounds[0] < bounds[2]
    assert bounds[1] < bounds[3]


def test_cells_bounds_arrays():
    h3indexes = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
            h3.geo_to_h3(10.3, 45.1, 5),
        ],
        dtype=np.uint64,
    )
    bounds_df = cells_bounds_arrays(h3indexes)
    assert bounds_df is not None
    assert isinstance(bounds_df, pl.DataFrame)
    assert len(bounds_df) == 2
    assert "minx" in bounds_df
    assert "maxx" in bounds_df
    assert "miny" in bounds_df
    assert "maxy" in bounds_df
    assert bounds_df["minx"][0] < 45.1
    assert bounds_df["maxx"][0] > 45.1
    assert bounds_df["miny"][0] < 10.3
    assert bounds_df["maxy"][0] > 10.3



from h3ronpy.arrow.vector import geometry_to_cells, ContainmentMode
import pyarrow as pa
import shapely
from shapely.geometry import Point
import h3.api.numpy_int as h3


def test_geometry_to_cells():
    geom = shapely.Polygon(((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)))
    cells = geometry_to_cells(geom, 5, containment_mode=ContainmentMode.IntersectsBoundary)
    assert isinstance(cells, pa.Array)
    assert cells.type == pa.uint64()
    assert len(cells) > 10


def test_geometry_to_cells_central_park():
    # Manhattan Central Park
    point = Point(-73.9575, 40.7938)

    arr = geometry_to_cells(point, 8).to_numpy()
    assert len(arr) == 1
    assert arr[0] == h3.geo_to_h3(point.y, point.x, 8)



import geopandas as gpd
import numpy as np
import pandas as pd
import typing

from ..arrow import raster as arrow_raster
from .vector import cells_dataframe_to_geodataframe

__doc__ = arrow_raster.__doc__

nearest_h3_resolution = arrow_raster.nearest_h3_resolution
rasterize_cells = arrow_raster.rasterize_cells


def raster_to_dataframe(
    in_raster: np.array,
    transform,
    h3_resolution: int,
    nodata_value=None,
    axis_order: str = "yx",
    compact: bool = True,
    geo: bool = False,
) -> typing.Union[gpd.GeoDataFrame, pd.DataFrame]:
    """
    Convert a raster/array to a pandas `DataFrame` containing H3 indexes

    This function is parallelized and uses the available CPUs by distributing tiles to a thread pool.

    The input geometry must be in WGS84.

    :param in_raster: input 2-d array
    :param transform:  the affine transformation
    :param nodata_value: the nodata value. For these cells of the array there will be no h3 indexes generated
    :param axis_order: axis order of the 2d array. Either "xy" or "yx"
    :param h3_resolution: target h3 resolution
    :param compact: Return compacted h3 indexes (see H3 docs). This results in mixed H3 resolutions, but also can
            reduce the amount of required memory.
    :param geo: Return a geopandas `GeoDataFrame` with geometries. increases the memory usage.
    :return: pandas `DataFrame` or `GeoDataFrame`
    """

    df = arrow_raster.raster_to_dataframe(
        in_raster, transform, h3_resolution, nodata_value=nodata_value, axis_order=axis_order, compact=compact
    ).to_pandas()

    if geo:
        return cells_dataframe_to_geodataframe(df)
    else:
        return df


def raster_to_geodataframe(*a, **kw) -> gpd.GeoDataFrame:
    """
    convert to a geodataframe

    Uses the same parameters as array_to_dataframe
    """
    kw["geo"] = True
    return raster_to_dataframe(*a, **kw)



import numpy as np
import h3.api.numpy_int as h3
from h3ronpy.polars import cells_to_string
import polars as pl


def some_cells() -> np.ndarray:
    return np.full(1000, h3.geo_to_h3(45.5, 10.2, 5), dtype="uint64")


def benchmark_h3_to_string_python_list(cells):
    return [h3.h3_to_string(cell) for cell in cells]


def test_cells_to_string(benchmark):
    benchmark(cells_to_string, pl.Series(some_cells()))


def test_h3_to_string_python_list(benchmark):
    benchmark(benchmark_h3_to_string_python_list, list(some_cells()))


h3_to_string_numpy_vectorized = np.vectorize(
    h3.h3_to_string,
    otypes=[
        str,
    ],
)


def test_h3_to_string_numpy_vectorized(benchmark):
    benchmark(h3_to_string_numpy_vectorized, some_cells())



from h3ronpy.arrow.raster import Transform


def test_transform_cmp():
    assert Transform(1, 1, 0, 1, 0, 1) == Transform(1, 1, 0, 1, 0, 1)
    assert Transform(1, 1, 0, 0, 0, 1) != Transform(1, 1, 0, 1, 0, 1)



"""
API to use `h3ronpy` with the `polars dataframe library <https://www.pola.rs/>`_.

.. warning::

    To avoid pulling in unused dependencies, `h3ronpy` does not declare a dependency to `polars`. This
    package needs to be installed separately.

"""

from functools import wraps
import polars as pl
import pyarrow as pa
from .. import arrow as _arrow


def _wrap(func, ret_type=None):
    @wraps(func, updated=())
    def wrapper(*args, **kw):
        result = func(*args, **kw)
        if isinstance(result, pa.Table) or isinstance(result, pa.Array):
            return pl.from_arrow(result)
        return result

    if ret_type:
        # create a copy to avoid modifying the dict of the wrapped function
        wrapper.__annotations__ = dict(**wrapper.__annotations__)
        wrapper.__annotations__["return"] = ret_type
    return wrapper


change_resolution = _wrap(_arrow.change_resolution, ret_type=pl.Series)
change_resolution_list = _wrap(_arrow.change_resolution, ret_type=pl.Series)
change_resolution_paired = _wrap(_arrow.change_resolution_paired, ret_type=pl.DataFrame)
cells_resolution = _wrap(_arrow.cells_resolution, ret_type=pl.Series)
cells_parse = _wrap(_arrow.cells_parse, ret_type=pl.Series)
vertexes_parse = _wrap(_arrow.vertexes_parse, ret_type=pl.Series)
directededges_parse = _wrap(_arrow.directededges_parse, ret_type=pl.Series)
compact = _wrap(_arrow.compact, ret_type=pl.Series)
uncompact = _wrap(_arrow.uncompact, ret_type=pl.Series)
cells_valid = _wrap(_arrow.cells_valid, ret_type=pl.Series)
vertexes_valid = _wrap(_arrow.vertexes_valid, ret_type=pl.Series)
directededges_valid = _wrap(_arrow.directededges_valid, ret_type=pl.Series)
grid_disk = _wrap(_arrow.grid_disk, ret_type=pl.Series)
grid_disk_distances = _wrap(_arrow.grid_disk_distances, ret_type=pl.DataFrame)
grid_ring_distances = _wrap(_arrow.grid_ring_distances, ret_type=pl.DataFrame)
grid_disk_aggregate_k = _wrap(_arrow.grid_disk_aggregate_k, ret_type=pl.DataFrame)
cells_area_m2 = _wrap(_arrow.cells_area_m2, ret_type=pl.Series)
cells_area_km2 = _wrap(_arrow.cells_area_km2, ret_type=pl.Series)
cells_area_rads2 = _wrap(_arrow.cells_area_rads2, ret_type=pl.Series)
cells_to_string = _wrap(_arrow.cells_to_string, ret_type=pl.Series)
vertexes_to_string = _wrap(_arrow.vertexes_to_string, ret_type=pl.Series)
directededges_to_string = _wrap(_arrow.directededges_to_string, ret_type=pl.Series)


@pl.api.register_expr_namespace("h3")
class H3Expr:
    """
    Registers H3 functionality with polars Expr expressions.

    The methods of this class mirror the functionality provided by the functions of this
    module. Please refer to the module functions for more documentation.
    """

    def __init__(self, expr: pl.Expr):
        self._expr = expr

    def cells_resolution(self) -> pl.Expr:
        return self._expr.map(lambda s: cells_resolution(s)).alias("resolution")

    def change_resolution(self, resolution: int) -> pl.Expr:
        return self._expr.map(lambda s: change_resolution(s, resolution))

    def change_resolution_list(self, resolution: int) -> pl.Expr:
        return self._expr.map(lambda s: change_resolution_list(s, resolution))

    def cells_parse(self, set_failing_to_invalid: bool = False) -> pl.Expr:
        return self._expr.map(lambda s: cells_parse(s, set_failing_to_invalid=set_failing_to_invalid)).alias("cell")

    def vertexes_parse(self, set_failing_to_invalid: bool = False) -> pl.Expr:
        return self._expr.map(lambda s: vertexes_parse(s, set_failing_to_invalid=set_failing_to_invalid)).alias(
            "vertex"
        )

    def directededges_parse(self, set_failing_to_invalid: bool = False) -> pl.Expr:
        return self._expr.map(lambda s: directededges_parse(s, set_failing_to_invalid=set_failing_to_invalid)).alias(
            "directededge"
        )

    def grid_disk(self, k: int, flatten: bool = False) -> pl.Expr:
        return self._expr.map(lambda s: grid_disk(s, k, flatten=flatten))

    def compact(self, mixed_resolutions: bool = False) -> pl.Expr:
        return self._expr.map(lambda s: compact(s, mixed_resolutions=mixed_resolutions))

    def uncompact(self, target_resolution: int) -> pl.Expr:
        return self._expr.map(lambda s: uncompact(s, target_resolution))

    def cells_area_m2(self) -> pl.Expr:
        return self._expr.map(lambda s: cells_area_m2(s)).alias("area_m2")

    def cells_area_km2(self) -> pl.Expr:
        return self._expr.map(lambda s: cells_area_km2(s)).alias("area_km2")

    def cells_area_rads2(self) -> pl.Expr:
        return self._expr.map(lambda s: cells_area_rads2(s)).alias("area_rads2")

    def cells_valid(self) -> pl.Expr:
        return self._expr.map(lambda s: cells_valid(s)).alias("cells_valid")

    def vertexes_valid(self) -> pl.Expr:
        return self._expr.map(lambda s: vertexes_valid(s)).alias("vertexes_valid")

    def directededges_valid(self) -> pl.Expr:
        return self._expr.map(lambda s: directededges_valid(s)).alias("directededges_valid")

    def cells_to_string(self) -> pl.Expr:
        return self._expr.map(lambda s: cells_to_string(s))

    def vertexes_to_string(self) -> pl.Expr:
        return self._expr.map(lambda s: vertexes_to_string(s))

    def directededges_to_string(self) -> pl.Expr:
        return self._expr.map(lambda s: directededges_to_string(s))


@pl.api.register_series_namespace("h3")
class H3SeriesShortcuts:
    """
    Registers H3 functionality with polars Series.

    The methods of this class mirror the functionality provided by the functions of this
    module. Please refer to the module functions for more documentation.
    """

    def __init__(self, s: pl.Series):
        self._s = s

    def cells_resolution(self) -> pl.Series:
        return cells_resolution(self._s)

    def change_resolution(self, resolution: int) -> pl.Series:
        return change_resolution(self._s, resolution)

    def change_resolution_list(self, resolution: int) -> pl.Series:
        return change_resolution_list(self._s, resolution)

    def cells_parse(self, set_failing_to_invalid: bool = False) -> pl.Series:
        return cells_parse(self._s, set_failing_to_invalid=set_failing_to_invalid)

    def vertexes_parse(self, set_failing_to_invalid: bool = False) -> pl.Series:
        return vertexes_parse(self._s, set_failing_to_invalid=set_failing_to_invalid)

    def directededges_parse(self, set_failing_to_invalid: bool = False) -> pl.Series:
        return directededges_parse(self._s, set_failing_to_invalid=set_failing_to_invalid)

    def grid_disk(self, k: int, flatten: bool = False) -> pl.Series:
        return grid_disk(self._s, k, flatten=flatten)

    def compact(self, mixed_resolutions: bool = False) -> pl.Series:
        return compact(self._s, mixed_resolutions=mixed_resolutions)

    def uncompact(self, target_resolution: int) -> pl.Series:
        return uncompact(self._s, target_resolution)

    def cells_area_m2(self) -> pl.Series:
        return cells_area_m2(self._s)

    def cells_area_km2(self) -> pl.Series:
        return cells_area_km2(self._s)

    def cells_area_rads2(self) -> pl.Series:
        return cells_area_rads2(self._s)

    def cells_valid(self) -> pl.Series:
        return cells_valid(self._s)

    def vertexes_valid(self) -> pl.Series:
        return vertexes_valid(self._s)

    def directededges_valid(self) -> pl.Series:
        return directededges_valid(self._s)

    def cells_to_string(self) -> pl.Series:
        return cells_to_string(self._s)

    def vertexes_to_string(self) -> pl.Series:
        return vertexes_to_string(self._s)

    def directededges_to_string(self) -> pl.Series:
        return directededges_to_string(self._s)


__all__ = [
    change_resolution.__name__,
    change_resolution_list.__name__,
    change_resolution_paired.__name__,
    cells_resolution.__name__,
    cells_parse.__name__,
    vertexes_parse.__name__,
    directededges_parse.__name__,
    compact.__name__,
    uncompact.__name__,
    cells_valid.__name__,
    vertexes_valid.__name__,
    directededges_valid.__name__,
    grid_disk.__name__,
    grid_disk_distances.__name__,
    grid_ring_distances.__name__,
    grid_disk_aggregate_k.__name__,
    cells_area_m2.__name__,
    cells_area_km2.__name__,
    cells_area_rads2.__name__,
    cells_to_string.__name__,
    vertexes_to_string.__name__,
    directededges_to_string.__name__,
    H3Expr.__name__,
    H3SeriesShortcuts.__name__,
]



try:
    import rasterio

    HAS_RASTERIO = True
except ImportError:
    # rasterio is an optional dependency
    HAS_RASTERIO = False

import numpy as np
import pytest

from h3ronpy.pandas.raster import raster_to_dataframe
from h3ronpy import DEFAULT_CELL_COLUMN_NAME

from tests import TESTDATA_PATH


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_r_tiff():
    dataset = rasterio.open(TESTDATA_PATH / "r.tiff")
    band = dataset.read(1)
    df = raster_to_dataframe(band, dataset.transform, 8, nodata_value=0, compact=True, geo=False)
    assert len(df) > 100
    assert df.dtypes[DEFAULT_CELL_COLUMN_NAME] == "uint64"
    assert df.dtypes["value"] == "uint8"


@pytest.mark.skipif(not HAS_RASTERIO, reason="requires rasterio")
def test_r_tiff_float32():
    dataset = rasterio.open(TESTDATA_PATH / "r.tiff")
    band = dataset.read(1).astype(np.float32)
    df = raster_to_dataframe(band, dataset.transform, 8, nodata_value=0.0, compact=True, geo=False)
    assert len(df) > 100
    assert df.dtypes[DEFAULT_CELL_COLUMN_NAME] == "uint64"
    assert df.dtypes["value"] == "float32"


def test_preserve_nan_without_nodata_value():
    arr = np.array([[np.nan, 1.0], [np.nan, 1.0]], dtype=np.float32)
    df = raster_to_dataframe(arr, [11.0, 1.0, 0.0, 10.0, 1.2, 0.2], 7, nodata_value=None)
    assert df["value"].value_counts(dropna=False)[np.nan] > 100
    assert df["value"].value_counts(dropna=False)[1.0] > 100


def test_preserve_nan_with_nodata_value():
    arr = np.array([[np.nan, 1.0], [np.nan, 1.0]], dtype=np.float32)
    df = raster_to_dataframe(arr, [11.0, 1.0, 0.0, 10.0, 1.2, 0.2], 7, nodata_value=1.0)
    assert df["value"].value_counts(dropna=False)[np.nan] > 100



from . import _wrap
from ..arrow import vector as _av
import polars as pl

cells_to_coordinates = _wrap(_av.cells_to_coordinates, ret_type=pl.DataFrame)
coordinates_to_cells = _wrap(_av.coordinates_to_cells, ret_type=pl.Series)
cells_bounds = _av.cells_bounds
cells_bounds_arrays = _wrap(_av.cells_bounds_arrays, ret_type=pl.DataFrame)
cells_to_wkb_polygons = _wrap(_av.cells_to_wkb_polygons, ret_type=pl.Series)
cells_to_wkb_points = _wrap(_av.cells_to_wkb_points, ret_type=pl.Series)
vertexes_to_wkb_points = _wrap(_av.vertexes_to_wkb_points, ret_type=pl.Series)
wkb_to_cells = _wrap(_av.wkb_to_cells, ret_type=pl.Series)
geometry_to_cells = _wrap(_av.geometry_to_cells, ret_type=pl.Series)

__all__ = [
    cells_to_coordinates.__name__,
    coordinates_to_cells.__name__,
    cells_bounds.__name__,
    cells_bounds_arrays.__name__,
    cells_to_wkb_polygons.__name__,
    cells_to_wkb_points.__name__,
    vertexes_to_wkb_points.__name__,
    wkb_to_cells.__name__,
    geometry_to_cells.__name__,
]



import numpy as np
import h3.api.numpy_int as h3
from h3ronpy.polars import cells_area_km2
import polars as pl


def test_cells_area_km2():
    cells = np.array(
        [
            h3.geo_to_h3(10.3, 45.1, 8),
            h3.geo_to_h3(10.3, 45.1, 5),
            h3.geo_to_h3(10.3, 45.1, 3),
        ],
        dtype=np.uint64,
    )
    areas = cells_area_km2(cells)
    assert isinstance(areas, pl.Series)
    assert len(areas) == 3
    assert int(areas[0] * 100) == 62
    assert int(areas[1]) == 213
    assert int(areas[2]) == 10456
